#include <windows.h>
#include <wtypes.h>
#include "wmsdk.h"
#include <stdio.h>
#include "resource.h"
#include "plugin.h"


static HINSTANCE hInstance = NULL;
static POINT position = {0, 0};
static DWORD enabled = FALSE;


BOOL APIENTRY DllMain( HANDLE hModule, DWORD reason, LPVOID resvd )
{
  hInstance = (HINSTANCE)hModule;
  return TRUE;
}


static void FormatString(char *in, char *fname, char *out, int len,
			 char *title, char *artist, char *album,
			 char *year, int track, char *bitrate, char *duration)
{
  int i, j, slen;
  char obuff[8192];
  char buff[8];
	
  if (in[0]==0) { strncpy(out, fname, len); return; }

  slen = strlen(in);
  for (i=j=0; i<slen && j<8000; i++) {
    obuff[j] = 0;
    if (!strncmp(in+i, "%title%", 7)) {
      strcat(obuff+j, title);
      i += 6;
      j += strlen(title);
    } else if (!strncmp(in+i, "%artist%", 8)) {
      strcat(obuff+j, artist);
      i += 7;
      j += strlen(artist);
    } else if (!strncmp(in+i, "%album%", 7)) {
      strcat(obuff+j, album);
      i += 6;
      j += strlen(album);
    } else if (!strncmp(in+i, "%year%", 6)) {
      strcat(obuff+j, year);
      i += 5;
      j += strlen(year);
    } else if (!strncmp(in+i, "%bitrate%", 9)) {
      strcat(obuff+j, bitrate);
      i += 8;
      j += strlen(bitrate);
    } else if (!strncmp(in+i, "%duration%", 10)) {
      strcat(obuff+j, duration);
      i += 9;
      j += strlen(duration);
    } else if (!strncmp(in+i, "%track%", 7)) {
      if (track>0)
	sprintf(buff, "%2d", track);
      else
	buff[0] = 0;
      strcat(obuff+j, buff);
      i += 6;
      j += strlen(buff);
    } else
      obuff[j++] = in[i];
  }
  obuff[j]=0;
  strncpy(out, obuff, len);
  out[len-1] = 0;
}

static void CenterDialog( HWND hwndDlg )
{
  HWND hwndOwner; 
  RECT rc, rcDlg, rcOwner; 

  hwndOwner = GetParent(hwndDlg);
  if (hwndOwner == NULL) hwndOwner = GetDesktopWindow(); 
  GetWindowRect(hwndOwner, &rcOwner); 
  GetWindowRect(hwndDlg, &rcDlg); 
  CopyRect(&rc, &rcOwner); 
  OffsetRect(&rcDlg, -rcDlg.left, -rcDlg.top); 
  OffsetRect(&rc, -rc.left, -rc.top); 
  OffsetRect(&rc, -rcDlg.right, -rcDlg.bottom); 
  SetWindowPos(hwndDlg, HWND_TOP, rcOwner.left + (rc.right / 2),
	       rcOwner.top + (rc.bottom / 2), 0, 0, SWP_NOSIZE); 
}

static void PositionWindow( HWND hdlg, POINT *pt)
{
  HDC hdc;
  if (pt->x!=0 && pt->y!=0) {
    hdc=CreateIC("DISPLAY", "", "", NULL);
    pt->x = min( GetDeviceCaps(hdc, HORZRES)-100, pt->x);
    pt->x = max( pt->x, 0 );
    pt->y = min( GetDeviceCaps(hdc, VERTRES)-100, pt->y);
    pt->y = max( pt->y, 0 );
    DeleteDC(hdc);
    SetWindowPos(hdlg, HWND_TOP, pt->x, pt->y, 0,0, SWP_NOSIZE);
  } else
    CenterDialog(hdlg);
}

static void SaveWindowPosition( HWND hdlg, POINT *pt )
{
  RECT saverect;
  GetWindowRect(hdlg, &saverect);
  pt->x = saverect.left;
  pt->y = saverect.top;
}

EXPORT int FileFilterInitialize( MINIFILEPARAM *param )
{
  DWORD type, len;

  // First, check that we can grok this plugin version!
  if (!param) return FALSE;
  if (param->version != PLUGINVERSION) return FALSE;

  strncpy(param->title, "OGG File Renamer", MAXTITLE);
  param->version = PLUGINVERSION;
  param->user = calloc(512, 1);
  strcpy((char*)param->user, "%title% - %artist%");
  strcpy((char*)param->user+256, "%duration%");

  len = 256;
  type = REG_SZ;
  RegQueryValueEx(param->userKey, "string", 0, &type,
		  (unsigned char *)param->user, &len);

  len = 256;
  type = REG_SZ;
  RegQueryValueEx(param->userKey, "commstring", 0, &type,
		  (unsigned char *)param->user+256, &len);

  type = REG_BINARY;
  len = sizeof(POINT);
  RegQueryValueEx(param->userKey, "position", 0, &type,
		  (unsigned char *)&position, &len);

  type = REG_DWORD;
  len = sizeof(DWORD);
  RegQueryValueEx(param->userKey, "enabled", 0, &type,
		  (unsigned char *)&enabled, &len);
	
  CoInitialize(NULL);
  
  return TRUE;
}


static BOOL CALLBACK Dialog(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  static MINIFILEPARAM *param;
  static int last = 0;
  char buff[256], outBuff[256];
  int en;

  switch(msg) {
  case WM_USER:
    en = (SendDlgItemMessage(hDlg, IDC_ENABLE, BM_GETCHECK, 0, 0)==BST_CHECKED)
      ?TRUE:FALSE;
    EnableWindow(GetDlgItem(hDlg, IDC_EDIT), en);
    EnableWindow(GetDlgItem(hDlg, IDC_EDITCOMM), en);
    EnableWindow(GetDlgItem(hDlg, IDC_TITLE), en);
    EnableWindow(GetDlgItem(hDlg, IDC_ALBUM), en);
    EnableWindow(GetDlgItem(hDlg, IDC_ARTIST), en);
    EnableWindow(GetDlgItem(hDlg, IDC_BITRATE), en);
    EnableWindow(GetDlgItem(hDlg, IDC_DURATION), en);
    EnableWindow(GetDlgItem(hDlg, IDC_YEAR), en);
    EnableWindow(GetDlgItem(hDlg, IDC_TRACK), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATIC1), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATIC2), en);
    EnableWindow(GetDlgItem(hDlg, IDC_SAMPLE), en);
    EnableWindow(GetDlgItem(hDlg, IDC_SAMPLECOMM), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATICN), en);
    EnableWindow(GetDlgItem(hDlg, IDC_STATICC), en);
    if (en) {
      GetDlgItemText(hDlg, IDC_EDIT, buff, 64);
      FormatString(buff, "Filename.ogg", outBuff, 256, "Song Title",
		   "The Ogglodytes", "Red Album", "2002", 3, "64Kbps", "3:36");
      SetDlgItemText(hDlg, IDC_SAMPLE, outBuff);

      GetDlgItemText(hDlg, IDC_EDITCOMM, buff, 64);
      FormatString(buff, "1.43MB", outBuff, 256, "Song Title", "The Ogglodytes",
		   "Red Album", "2002", 3, "128Kbps", "3:36");
      SetDlgItemText(hDlg, IDC_SAMPLECOMM, outBuff);
    } else {
      SetDlgItemText(hDlg, IDC_SAMPLE, "Filename.ogg");
      SetDlgItemText(hDlg, IDC_SAMPLECOMM, "1.43MB");
    }
    return TRUE;

  case WM_INITDIALOG:
    param = (MINIFILEPARAM*)lParam;
    PositionWindow(hDlg, &position);
    SendDlgItemMessage(hDlg, IDC_ENABLE, BM_SETCHECK, 
		       enabled?BST_CHECKED:BST_UNCHECKED, 0);
    SetDlgItemText(hDlg, IDC_EDIT, (char*)param->user);
    SetDlgItemText(hDlg, IDC_EDITCOMM, (char*)param->user+256);
    SetFocus(GetDlgItem(hDlg, enabled?IDC_EDIT:IDC_ENABLE));
    SendMessage(hDlg, WM_USER, 0, 0);
    last = IDC_EDIT;
    return FALSE;


  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDOK:
      SaveWindowPosition(hDlg, &position);
      GetDlgItemText(hDlg, IDC_EDIT, (char*)param->user, 256);
      GetDlgItemText(hDlg, IDC_EDITCOMM, ((char*)param->user)+256, 256);
      if (SendDlgItemMessage(hDlg, IDC_ENABLE, BM_GETCHECK, 0, 0)==BST_CHECKED)
	enabled = TRUE;
      else
	enabled = FALSE;
      EndDialog(hDlg, TRUE);
      return TRUE;
		
    case IDCANCEL:
      SaveWindowPosition(hDlg, &position);
      EndDialog(hDlg, FALSE);
      return TRUE;
		
    case IDC_EDIT:
      last = IDC_EDIT;
      return SendMessage(hDlg, WM_USER, 0, 0);

    case IDC_EDITCOMM:
      last = IDC_EDITCOMM;
      return SendMessage(hDlg, WM_USER, 0, 0);

    case IDC_ENABLE:
      SendMessage(hDlg, WM_USER, 0, 0);
      return TRUE;

    case IDC_TITLE:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%title%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_ARTIST:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%artist%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_ALBUM:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%album%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_YEAR:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%year%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_TRACK:
      if (last!=IDC_EDITCOMM) SendDlgItemMessage(hDlg, last, EM_REPLACESEL,
						 FALSE, (long)"%track%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_BITRATE:
      SendDlgItemMessage(hDlg, last, EM_REPLACESEL, FALSE, (long)"%bitrate%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    case IDC_DURATION:
      SendDlgItemMessage(hDlg, last, EM_REPLACESEL, FALSE, (long)"%duration%");
      SetFocus(GetDlgItem(hDlg, last));
      return TRUE;

    }
    break;
  }
  return FALSE;
}



EXPORT void FileFilterUserInterface( MINIFILEPARAM *param, HWND parent )
{
  DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_DIALOG), parent, 
		 Dialog, (LPARAM)param);
}


static int CheckOGG(char *fname, char *title, char *artist, char *album,
		    char *year, int *trackI, char *bitrate, char *duration)
{
  FILE *fp;
  unsigned char buff[64], a, *hdr, *cmt;
  unsigned int blocks, size, i;
  unsigned long fsize;
  unsigned int br, sr;
  unsigned int ptr, tags;
  unsigned int hlen;
  char temptag[256];
  unsigned int dur;
  unsigned char bigBuff[16384];

  fp = fopen(fname, "rb");
  if (!fp) return FALSE;

  title[0] = artist[0] = album[0] = year[0] = bitrate[0] =
    duration[0] = *trackI = 0;

  buff[0] = 0;
  fread(buff, 27, 1, fp);
  if (!(buff[0]=='O' && buff[1]=='g' && buff[2]=='g' && buff[3]=='S')) {
    fclose(fp);
    return FALSE;
  }
  blocks = buff[26];
  size = 0;
  for (i=0; i<blocks; i++) {
    a = fgetc(fp);
    size += a;
    if (a!=0xff) {i++; break;}
  }
  for (;i<blocks; i++) fgetc(fp);
  hdr = (unsigned char *)malloc(size);
  if (!hdr) {
    fclose(fp);
    return FALSE;
  }
  fread(hdr, size, 1, fp);

  buff[0] = 0;
  fread(buff, 27, 1, fp);
  if (!(buff[0]=='O' && buff[1]=='g' && buff[2]=='g' && buff[3]=='S')) {
    fclose(fp);
    return FALSE;
  }
  blocks = buff[26];
  size = 0;
  for (i=0; i<blocks; i++) {
    a = fgetc(fp);
    size += a;
    if (a!=0xff) {i++; break;}
  }
  for (;i<blocks; i++) fgetc(fp);
  cmt = (unsigned char *)malloc(size);
  if (!cmt) {
    free(hdr);
    fclose(fp);
    return FALSE;
  }
  fread(cmt, size, 1, fp);

  memset(bigBuff, 0, 16384);
  fseek(fp, 0, SEEK_END);
  fseek(fp, max(0, ftell(fp)-16384), SEEK_SET);
  fread(bigBuff, 16384, 1, fp);
  fsize = ftell(fp);
  fclose(fp);

  if (hdr[0]!=0x01 || memcmp("vorbis", (char*)(hdr+1), 6) ||
      cmt[0]!=0x03 || memcmp("vorbis", (char*)(cmt+1), 6)    ) {
    free(hdr);
    free(cmt);
    return FALSE;
  }
	
  br = hdr[20]+(hdr[21]<<8);
  sr = hdr[12]+(hdr[13]<<8);

  // default to fsize/bitrate, not bad guesstimate
  fsize = ((fsize*8)/br)*sr;
  // Search backwards for last OggS header w/sample count...
  for (i=16384-32; i>0; i--) {
    if (bigBuff[i]=='O' && bigBuff[i+1]=='g'
	&& bigBuff[i+2]=='g' && bigBuff[i+3]=='S') {
      fsize = bigBuff[i+6]+(bigBuff[i+7]<<8)+(bigBuff[i+8]<<16)
	+(bigBuff[i+9]<<24);
      break;
    }
  }



  hlen = cmt[7]+(cmt[8]<<8)+(cmt[9]<<16)+(cmt[10]<<24);
  tags = cmt[11+hlen] + (cmt[12+hlen]<<8) + (cmt[13+hlen]<<16) +
    (cmt[14+hlen]<<24);
  ptr = 15+hlen;
  for (i=0; i<tags; i++) {
    size = cmt[ptr] + (cmt[ptr+1]<<8) + (cmt[ptr+2]<<16) + (cmt[ptr+3]<<24);
    strncpy(temptag, (char*)(cmt+ptr+4), min(size, 255));
    temptag[min(size, 255)] = 0;
    ptr += size + 4;

    if (!strnicmp(temptag, "TITLE=", 6)) strncpy(title, temptag+6, 64);
    else if (!strnicmp(temptag, "ARTIST=", 7)) strncpy(artist, temptag+7, 64);
    else if (!strnicmp(temptag, "ALBUM=", 6)) strncpy(album, temptag+6, 64);
    else if (!strnicmp(temptag, "YEAR=", 5)) strncpy(year, temptag+5, 64);
    else if (!strnicmp(temptag, "TRACKNUMBER=", 12))
      sscanf(temptag+12, "%d", trackI);
  }
  free(hdr);
  free(cmt);
  sprintf(bitrate, "%dKbps", (br+1023)/1024);
  dur = fsize/sr;
  sprintf(duration, "%0d:%02d", dur/60, dur%60);
  return TRUE;
}


EXPORT MINIZIPINFO *FileFilterExecute( MINIFILEPARAM *param,
				       MINIZIPINFO *mz,
				       char *label,
				       unsigned int labelSize,
				       char *path )
{
  int track;
  unsigned int i;
  char fname[MAX_PATH], buff[1024], title[128], artist[128];
  char album[128], year[128], bitrate[128], duration[128];
  char outBuff[1024];
  char *fileMap, *commMap, *dot;

  if (!enabled)
    return mz;

  fileMap = (char *)param->user;
  commMap = fileMap + 256;

  for (i=0; i<mz->numfiles; i++) {
		dot = strrchr(mz->fi[i].name, '.');
		if (!dot) dot = "x";
    if (!stricmp(dot, ".ogg")) {
      if (path[strlen(path)-1]=='\\')
	sprintf(fname, "%s%s", path, mz->fi[i].name);
      else
	sprintf(fname, "%s\\%s", path, mz->fi[i].name);
      buff[0] = 0;

      if (CheckOGG(fname, title, artist, album, year,
		   &track, bitrate, duration)) {
	if (!title[0] || (title[0]==' '&&title[1]==0)) {
	  // No renaming...
	} else {
	  FormatString(fileMap, mz->fi[i].name, outBuff, 256,
		       title, artist, album, year, track,
		       bitrate, duration);
	  strncpy(mz->fi[i].name, outBuff, NAMELEN);
	  mz->fi[i].name[NAMELEN-1]=0;

	  FormatString(commMap, mz->fi[i].comm, outBuff, 256,
		       title, artist, album, year, track,
		       bitrate, duration);
	  strncpy(mz->fi[i].comm, outBuff, COMMLEN);
	  mz->fi[i].comm[COMMLEN-1]=0;
	}
      }    
    }

    if (mz->fi[i].zi) {
      sprintf(fname, "%s\\%s", path, mz->fi[i].name);
      FileFilterExecute( param, mz->fi[i].zi, label, labelSize, fname);
    }
  }
  return mz;
}

EXPORT void FileFilterUninitialize( MINIFILEPARAM *param )
{
  RegSetValueEx(param->userKey, "string", 0, REG_SZ,
		(unsigned char*)param->user, 256);
  RegSetValueEx(param->userKey, "commstring", 0, REG_SZ,
		(unsigned char*)param->user+256, 256);
  RegSetValueEx(param->userKey, "position", 0, REG_BINARY,
		(unsigned char*)&position, sizeof(POINT));
  RegSetValueEx(param->userKey, "enabled", 0, REG_DWORD,
		(unsigned char*)&enabled, sizeof(DWORD));
  free(param->user);
}


